import re
import logging
from typing import Any, Optional

from .dataclass import Field, XedaBaseModel, root_validator

log = logging.getLogger(__name__)


__all__ = [  #
    "FPGA",  #
]


class FPGA(XedaBaseModel):
    """FPGA target device"""

    # definition order: part > device > vendor > {family, speed, package, etc}
    part: Optional[str] = Field(None, description="full device part identifier")
    device: Optional[str]
    vendor: Optional[str]
    family: Optional[str]
    generation: Optional[str]
    type: Optional[str]
    speed: Optional[str] = Field(None, description="speed-grade")
    package: Optional[str]
    capacity: Optional[str]
    pins: Optional[int]
    grade: Optional[str]

    def __init__(self, *args: str, **data: Any) -> None:
        if args:
            if len(args) != 1 or not args[0]:
                raise ValueError("Only a single 'part' non-keyword argument is valid.")
            a = args[0]
            if isinstance(a, str):
                if "part" in data:
                    raise ValueError("'part' field already given in keyword arguments.")
                data["part"] = a
            elif isinstance(a, dict):
                if data:
                    raise ValueError("Both dictionary and keyword arguments preset")
                data = a
            else:
                raise ValueError(f"Argument of type {type(a)} is not supported!")
        log.debug("fpga init! data=%s", data)
        super().__init__(**data)

    # this is called before all field validators!
    @root_validator(pre=True)
    def _fpga_root_validator(cls, values):  # pylint: disable=no-self-argument
        if not values:
            return values
        # Intel: https://www.intel.com/content/dam/www/central-libraries/us/en/documents/product-catalog.pdf
        # Lattice: https://www.latticesemi.com/Support/PartNumberReferenceGuide
        # Xilinx: https://www.xilinx.com/support/documents/selection-guides/7-series-product-selection-guide.pdf
        #         https://docs.xilinx.com/v/u/en-US/ds890-ultrascale-overview
        #         https://www.xilinx.com/support/documents/selection-guides/ultrascale-fpga-product-selection-guide.pdf
        #         https://www.xilinx.com/support/documents/selection-guides/ultrascale-plus-fpga-product-selection-guide.pdf
        part = values.get("part")

        def set_if_not_exist(attr: str, v: Any) -> None:
            if attr not in values:
                values[attr] = v

        def set_xc_family(s: str):
            d = dict(s="spartan", a="artix", k="kintex", v="virtex", z="zynq")
            s = s.lower()
            if s in d:
                set_if_not_exist("family", d[s])

        if part:
            part = part.strip()
            values["part"] = part
            # speed: 6 = slowest, 8 = fastest
            match_ecp5 = re.match(
                r"^LFE5(U|UM|UM5G)-(\d+)F-(?P<sp>\d)(?P<pkg>[A-Z]+)(?P<pins>\d+)(?P<gr>[A-Z]?)$",
                part,
                flags=re.IGNORECASE,
            )
            if match_ecp5:
                log.debug(
                    "Lattice ECP5 attributes extracted from part# (%s): %s",
                    part,
                    match_ecp5.groupdict(),
                )
                set_if_not_exist("vendor", "lattice")
                set_if_not_exist("family", "ecp5")
                set_if_not_exist("type", match_ecp5.group(1).lower())
                set_if_not_exist("device", "LFE5" + match_ecp5.group(1).upper())
                set_if_not_exist("capacity", match_ecp5.group(2) + "k")
                set_if_not_exist("speed", match_ecp5.group("sp"))
                set_if_not_exist("package", match_ecp5.group("pkg"))
                set_if_not_exist("pins", int(match_ecp5.group("pins")))
                set_if_not_exist("grade", match_ecp5.group("gr"))
                return values
            # Commercial Xilinx # Generation # Family # Logic Cells in 1K units # Speed Grade (-1 slowest, L: low-power) # Package Type
            match_xc7 = re.match(
                r"^(XC)(?P<g>\d)(?P<f>[A-Z])(?P<lc>\d+)-?(?P<s>L?\d)?(?P<pkg>[A-Z]+)(?P<pins>\d+)(?P<gr>-\d)?$",
                part,
                flags=re.IGNORECASE,
            )
            if match_xc7:
                log.debug(
                    "Xilinx/AMD series-7 attributes extracted from part# (%s): %s",
                    part,
                    match_xc7.groupdict(),
                )
                set_if_not_exist("vendor", "xilinx")
                set_if_not_exist("generation", match_xc7.group("g"))
                set_xc_family(match_xc7.group("f") + "-7")
                lc = match_xc7.group("lc")
                set_if_not_exist(
                    "device",
                    match_xc7.group(1)
                    + match_xc7.group("g")
                    + match_xc7.group("f")
                    + lc,
                )
                set_if_not_exist("capacity", lc + "K")
                set_if_not_exist("package", match_xc7.group("pkg"))
                set_if_not_exist("pins", int(match_xc7.group("pins")))
                set_if_not_exist("grade", match_xc7.group("gr"))
                return values
            match_us = re.match(
                r"^(XC)(?P<f>[A-Z])(?P<g>[A-Z]+)(?P<lc>\d+)-?(?P<s>L?\d)?(?P<pkg>[A-Z][A-Z][A-Z]+)(?P<pins>\d+)(?P<gr>-\d)?$",
                part,
                flags=re.IGNORECASE,
            )
            if match_us:
                log.debug(
                    "Xilinx/AMD Ultrascale attributes extracted from part# (%s): %s",
                    part,
                    match_us.groupdict(),
                )
                set_if_not_exist("vendor", "xilinx")
                set_if_not_exist("generation", match_us.group("g"))
                set_xc_family(match_us.group("f") + "-us")
                lc = match_us.group("lc")
                set_if_not_exist(
                    "device",
                    match_us.group(1) + match_us.group("g") + match_us.group("f") + lc,
                )
                set_if_not_exist("capacity", lc + "K")
                set_if_not_exist("package", int(match_us.group("pkg")))
                set_if_not_exist("pins", int(match_us.group("pins")))
                set_if_not_exist("grade", match_us.group("gr"))
                return values
            # UltraSCALE+
            # capacity is index to table, roughly x100K LCs
            match_usp = re.match(
                r"^(XC)(?P<f>[A-Z])U(?P<lc>\d+)P-?(?P<s>L?\d)?(?P<pkg>[A-Z]+)(?P<pins>\d+)(?P<gr>-\d)?$",
                part,
                flags=re.IGNORECASE,
            )
            if match_usp:
                log.debug(
                    "Xilinx/AMD Ultrascale+ attributes extracted from part# (%s): %s",
                    part,
                    match_usp.groupdict(),
                )
                set_if_not_exist("vendor", "xilinx")
                set_if_not_exist("generation", "usp")
                set_xc_family(match_usp.group("f") + "-usp")
                lc = match_usp.group("lc")
                set_if_not_exist("capacity", lc)
                set_if_not_exist(
                    "device", match_usp.group(1) + match_usp.group("f") + "U" + lc + "P"
                )
                set_if_not_exist("package", int(match_usp.group("pkg")))
                set_if_not_exist("pins", int(match_usp.group("pins")))
                set_if_not_exist("grade", match_usp.group("gr"))
                return values
        elif not values.get("device") and not values.get("vendor"):
            raise ValueError(
                "Missing enough information about the FPGA device. Please set the 'part' number and/or device, vendor, family, etc."
            )
        return values



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Dict, List, Optional, Union

import attrs
import yaml

from .design import Design
from .utils import WorkingDirectory, tomllib


@attrs.define
class XedaProject:
    # validate to concrete Designs to verify the whole xedaproject
    designs: List[Design]
    # keep raw dict as flows are dynamically discovered
    flows: Dict[str, dict]  # = attrs.field(default={}, validator=type_validator())

    @classmethod
    def from_file(cls, file: Union[str, os.PathLike, Path]):
        """load xedaproject from file"""
        if not isinstance(file, Path):
            file = Path(file)
        ext = file.suffix.lower()
        with open(file, "rb" if ext == ".toml" else "r") as f:
            if ext == ".toml":
                data = tomllib.load(f)
            elif ext == ".json":
                data = json.load(f)
            elif ext == ".yaml":
                data = yaml.safe_load(f)
            else:
                raise ValueError(
                    f"File {file} has unknown extension {ext}. Supported formats are TOML, JSON, and YAML."
                )
        if not isinstance(data, dict) or not data:
            raise ValueError("Invalid xedaproject!")
        designs = data.get("design") or data.get("designs")
        if not isinstance(designs, list):
            designs = [designs]
        if not designs:
            raise ValueError("No designs found in the xedaproject file!")

        flows = data.get("flow") or data.get("flows", {})
        assert isinstance(flows, dict)
        with WorkingDirectory(file.parent):
            return cls(  # type: ignore
                designs=[Design(**d) for d in designs if isinstance(d, dict)],
                flows=flows,
            )

    @property
    def design_names(self) -> List[str]:
        return [d.name for d in self.designs]

    def get_design(self, name: Optional[str] = None) -> Optional[Design]:
        if name is None:
            return self.designs[0] if len(self.designs) == 1 else None
        try:
            return self.designs[self.design_names.index(name)]
        except ValueError:
            return None



import logging
from typing import Optional, List, Union
from urllib.request import urlretrieve
from urllib.parse import urlparse
from urllib.error import HTTPError

from .yosys import YosysSynth
from .flow import FlowFatalError, FpgaSynthFlow
from ..tool import Tool
from ..utils import setting_flag
from ..dataclass import Field, XedaBaseModel, validator
from ..board import WithFpgaBoardSettings, get_board_file_path, get_board_data

__all__ = ["Nextpnr"]

log = logging.getLogger(__name__)


class EcpPLL(Tool):
    class Clock(XedaBaseModel):
        name: Optional[str] = None
        mhz: float = Field(gt=0, description="frequency in MHz")
        phase: Optional[float] = None

    executable = "ecppll"
    module: str = "ecp5pll"
    reset: bool = False
    standby: bool = False
    highres: bool = False
    internal_feedback: bool = False
    internal_feedback_wake: bool = False
    clkin: Union[Clock, float] = 25.0
    clkouts: List[Union[Clock, float]]
    file: Optional[str] = None

    @classmethod
    def _fix_clock(cls, clock, out_clk=None) -> Clock:
        if isinstance(clock, float):
            clock = cls.Clock(mhz=clock)
        if not clock.name:
            clock.name = "clk_i" if out_clk is None else f"clk_o_{out_clk}"
        return clock

    @validator("clkin", always=True)  # type: ignore
    def _validate_clockin(cls, value):
        return cls._fix_clock(value)

    @validator("clkouts", always=True)  # type: ignore
    def _validate_clockouts(cls, value):
        new_value = []
        for i, v in enumerate(value):
            new_value.append(cls._fix_clock(v, i))
        return new_value

    @validator("file", always=True)
    def _validate_outfile(cls, value, values):
        if not value:
            value = values["module"] + ".v"
        return value

    def generate(self):
        def s(f: float) -> str:
            return f"{f:0.03f}"

        args = setting_flag(self.module)
        args += setting_flag(self.file)
        args += ["--clkin_name", self.clkin.name]  # type: ignore # pylint: disable=no-member
        args += ["--clkin", s(self.clkin.mhz)]  # type: ignore # pylint: disable=no-member
        if not 1 <= len(self.clkouts) <= 4:
            raise ValueError("At least 1 and at most 4 output clocks can be specified")
        for i, clk in enumerate(self.clkouts):
            assert isinstance(clk, self.Clock) and clk.name
            args += [f"--clkout{i}_name", clk.name]
            args += [f"--clkout{i}", s(clk.mhz)]
            if clk.phase:
                if i > 0:
                    args += [f"--phase{i}", s(clk.phase)]
                else:
                    raise ValueError(
                        "First output clock cannot have a phase difference!"
                    )
        args += setting_flag(self.highres)
        args += setting_flag(self.standby)
        args += setting_flag(self.internal_feedback)
        self.run(*args)


class Nextpnr(FpgaSynthFlow):
    class Settings(WithFpgaBoardSettings):
        verbose: bool = False
        lpf_cfg: Optional[str] = None
        seed: Optional[int] = None
        randomize_seed: bool = False
        timing_allow_fail: bool = False
        ignore_loops: bool = Field(
            False, description="ignore combinational loops in timing analysis"
        )

        textcfg: Optional[str] = "config.txt"
        out_of_context: bool = Field(
            False,
            description="disable IO buffer insertion and global promotion/routing, for building pre-routed blocks",
        )
        lpf_allow_unconstrained = Field(
            False,
            description="don't require LPF file(s) to constrain all IOs",
        )
        extra_args: List[str] = []
        py_script: Optional[str] = None
        write: Optional[str] = None
        sdf: Optional[str] = None
        log: Optional[str] = "nextpnr.log"
        report: Optional[str] = "report.json"
        detailed_timing_report: bool = False  # buggy and likely to segfault
        placed_svg: Optional[str] = None  # "placed.svg"
        routed_svg: Optional[str] = None  # "routed.svg"
        parallel_refine: bool = False
        yosys: Optional[YosysSynth.Settings] = None

        @validator("yosys", always=True, pre=False)
        def _validate_yosys(cls, value, values):
            clocks = values.get("clocks")
            fpga = values.get("fpga")
            if value is None:
                value = YosysSynth.Settings(
                    fpga=fpga,
                    clocks=clocks,
                )  # pyright: reportGeneralTypeIssues=none
            else:
                if not isinstance(value, YosysSynth.Settings):
                    value = YosysSynth.Settings(**value)
                value.fpga = fpga
                value.clocks = clocks
            return value

    def init(self) -> None:
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        assert ss.yosys is not None
        self.add_dependency(
            YosysSynth,
            ss.yosys,
        )

    def run(self) -> None:
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        yosys_flow = self.completed_dependencies[0]
        assert isinstance(yosys_flow, YosysSynth)
        assert isinstance(yosys_flow.settings, YosysSynth.Settings)
        assert yosys_flow.settings.netlist_json
        netlist_json = yosys_flow.run_path / yosys_flow.settings.netlist_json
        fpga_family = ss.fpga.family if ss.fpga.family else "generic"
        assert fpga_family in {
            "generic",
            "ecp5",
            "ice40",
            "nexus",
            "gowin",
            "fpga-interchange",
            "xilinx",
        }, "unsupported fpga family"
        next_pnr = Tool(f"nextpnr-{fpga_family}")

        if not netlist_json.exists():
            raise FlowFatalError(f"netlist json file {netlist_json} does not exist!")

        lpf = ss.lpf_cfg
        board_data = get_board_data(ss.board)
        if not lpf and board_data and "lpf" in board_data:
            uri = board_data["lpf"]
            r = urlparse(uri)
            if r.scheme and r.netloc:
                try:
                    lpf, _ = urlretrieve(uri)
                except HTTPError as e:
                    log.critical(
                        "Unable to retrive file from %s (HTTP Error %d)",
                        uri,
                        e.code,
                    )
                    raise FlowFatalError("Unable to retreive LPF file") from None
            else:
                if "name" in board_data:
                    board_name = board_data["name"]
                else:
                    board_name = ss.board
                lpf = get_board_file_path(f"{board_name}.lpf")

        args = setting_flag(netlist_json, name="json")
        args += setting_flag(ss.clock_period and (1000 / ss.clock_period), name="freq")
        args += setting_flag(lpf)
        args += setting_flag(self.design.rtl.top)
        args += setting_flag(ss.seed)
        args += setting_flag(ss.fpga.speed)
        if ss.fpga.capacity:
            device_type = ss.fpga.type
            if device_type and device_type.lower() != "u":
                args.append(f"--{device_type}-{ss.fpga.capacity}")
            else:
                args.append(f"--{ss.fpga.capacity}")
        args += setting_flag(ss.out_of_context)
        args += setting_flag(ss.lpf_allow_unconstrained)
        args += setting_flag(ss.debug)
        args += setting_flag(ss.verbose)
        args += setting_flag(ss.quiet)
        args += setting_flag(ss.randomize_seed)
        args += setting_flag(ss.timing_allow_fail)
        args += setting_flag(ss.ignore_loops)
        args += setting_flag(ss.py_script, name="run")
        package = ss.fpga.package
        if ss.fpga.vendor and ss.fpga.vendor.lower() == "lattice":
            if package:
                package = package.upper()
                if package == "BG":
                    package = "CABGA"
                elif package == "MG":
                    package = "CSFBGA"
                assert ss.fpga.pins
                package += str(ss.fpga.pins)
        args += setting_flag(package)
        args += setting_flag(lpf)
        args += setting_flag(ss.textcfg)
        args += setting_flag(ss.write)
        args += setting_flag(ss.nthreads, name="threads")
        args += setting_flag(ss.sdf)
        args += setting_flag(ss.log)
        args += setting_flag(ss.report)
        args += setting_flag(ss.placed_svg)
        args += setting_flag(ss.routed_svg)
        args += setting_flag(ss.detailed_timing_report)
        args += setting_flag(ss.parallel_refine)
        if ss.extra_args:
            args += ss.extra_args
        next_pnr.run(*args)



#!/usr/bin/env python3
import logging
from xeda.flows.nextpnr import EcpPLL

log = logging.getLogger(__name__)

log.root.setLevel(logging.DEBUG)

f_pixel = 25.0

if __name__ == "__main__":
    EcpPLL(clkouts=[600]).generate()  # type: ignore



# © 2022 [Kamyar Mohajerani](mailto:kamyar@ieee.org)
"""Xeda Command-line interface"""
from __future__ import annotations

import inspect
import logging
import multiprocessing
import os
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Optional, Tuple

import click
import coloredlogs
from click.shell_completion import get_completion_class
from rich import box
from rich.style import Style
from rich.table import Table
from simple_term_menu import TerminalMenu
from typeguard.importhook import install_import_hook

from .cli_utils import (
    ClickMutex,
    OptionEatAll,
    XedaHelpGroup,
    XedaOptions,
    discover_flow_class,
    print_flow_settings,
    settings_to_dict,
)
from .console import console
from .design import Design, DesignValidationError
from .flow_runner import DefaultRunner
from .flows.flow import (
    Flow,
    FlowException,
    FlowFatalError,
    FlowSettingsError,
    registered_flows,
)
from .tool import ExecutableNotFound, NonZeroExitCode
from .xedaproject import XedaProject

install_import_hook("xeda")

log = logging.getLogger(__name__)


def load_design_from_toml(design_file) -> Design:
    return Design.from_toml(design_file)


def get_available_flows():
    """alternative method using inspect"""
    mod = "xeda.flows"
    fc = inspect.getmembers(
        sys.modules[mod],
        lambda cls: inspect.isclass(cls)
        and issubclass(cls, Flow)
        and not inspect.isabstract(cls),
    )
    return {n: (mod, cls) for n, cls in fc}


CONTEXT_SETTINGS = dict(
    auto_envvar_prefix="XEDA",
    help_option_names=["--help", "-h"],
    max_content_width=console.width,
)


def log_to_file(logdir: Path):
    timestamp = datetime.now().strftime("%Y-%m-%d-%H%M%S%f")[:-3]
    logdir.mkdir(exist_ok=True, parents=True)
    logfile = logdir / f"xeda_{timestamp}.log"
    log.info("Logging to %s", logfile)
    fileHandler = logging.FileHandler(logfile)
    logFormatter = logging.Formatter(
        "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s"
    )
    fileHandler.setFormatter(logFormatter)
    log.root.addHandler(fileHandler)


@click.group(cls=XedaHelpGroup, no_args_is_help=True, context_settings=CONTEXT_SETTINGS)
@click.option("--verbose", is_flag=True, help="Enables verbose mode.")
@click.option("--quiet", is_flag=True, help="Enable quiet mode.")
@click.option("--debug", show_envvar=True, is_flag=True)
@click.version_option(message="Xeda v%(version)s")
@click.pass_context
def cli(ctx: click.Context, **kwargs):
    ctx.obj = XedaOptions(**kwargs)
    log_level = (
        logging.WARNING
        if ctx.obj.quiet
        else logging.DEBUG
        if ctx.obj.debug
        else logging.INFO
    )
    log.root.setLevel(log_level)

    coloredlogs.install(
        None, fmt="%(asctime)s %(levelname)s %(message)s", logger=log.root
    )


@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Run a flow.",
    help="Run the flow identified by FLOW_NAME. A snake_case styled FLOW_NAME (e.g. ghdl_sim) is converted to a CamelCase class name (e.g. GhdlSim).",
    no_args_is_help=False,
)
@click.argument(
    "flow", metavar="FLOW_NAME", type=click.Choice(sorted(list(registered_flows)))
)
@click.option(
    "--xeda-run-dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
        allow_dash=True,
        path_type=Path,
    ),
    envvar="XEDA_RUN_DIR",
    help="Parent folder for execution of xeda commands.",
    default="xeda_run",
    show_default=True,
    show_envvar=True,
)
@click.option(
    "--cached-dependencies",
    default=True,
    help="Don't run dependency flows if a previous successfull run on the same design and flow settings exists. Generated directory names will contain a hash of design and/or flow settings.",
)
@click.option(
    "--run_in_existing_dir",
    is_flag=True,
    help="DO NOT USE!",
    hidden=True,
)
# @click.option(
#     "--force-run",
#     is_flag=True,
#     help="Force re-run of flow and all dependencies, even if they are already up-to-date",
# )
@click.option(
    "--xedaproject",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        allow_dash=False,
        path_type=Path,
    ),
    # cls=ClickMutex,
    # mutually_exclusive_with=["design_file"],
    help="Path to Xeda project file.",
)
@click.option(
    "--design-name",
    # cls=ClickMutex,
    # mutually_exclusive_with=["design_file"],
    help="Specify design.name in case multiple designs are available in a xedaproject.",
)
@click.option(
    "--design-file",
    "--design",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
        allow_dash=False,
        path_type=Path,
    ),
    # cls=ClickMutex,
    # mutually_exclusive_with=["xedaproject"],
    help="Path to Xeda design file containing the description of a single design.",
)
@click.option(
    "--flow-settings",
    "--settings",
    metavar="KEY=VALUE...",
    type=tuple,
    cls=OptionEatAll,
    help="""Override setting values for the executed flow. Separate multiple KEY=VALUE overrides with commas. KEY can be a hierarchical name using dot notation.
    Example: --settings clock_period=2.345 impl.strategy=Debug
    """
    #  examples: # FIXME move to docs
    # - xeda vivado_sim --flow-settings stop_time=100us
    # - xeda vivado_synth --flow-settings impl.strategy=Debug --flow-settings clock_period=2.345
)  # pylint: disable=C0116:missing-function-docstring
@click.pass_context
def run(
    ctx: click.Context,
    flow: str,
    cached_dependencies: bool,
    run_in_existing_dir: bool = False,
    # force_run: bool = False,
    xeda_run_dir: Optional[Path] = None,
    xedaproject: Optional[str] = None,
    design_name: Optional[str] = None,
    design_file: Optional[str] = None,
    flow_settings: Optional[Tuple[str, ...]] = None,
):
    assert ctx
    options: XedaOptions = ctx.obj or XedaOptions()
    assert xeda_run_dir
    log_to_file(xeda_run_dir / "Logs")
    # get default flow configs from xedaproject even if a design-file is specified
    xeda_project = None
    flows_config = {}
    if not xedaproject:
        xedaproject = "xedaproject.toml"
    if Path(xedaproject).exists():
        try:
            xeda_project = XedaProject.from_file(xedaproject)
        except DesignValidationError as e:
            log.critical("%s", e)
            sys.exit(1)
        except FileNotFoundError:
            sys.exit(
                f"Cannot open project file: {xedaproject}. Try specifing the correct path using the --xedaproject <path-to-file>."
            )
        flows_config = xeda_project.flows
    if design_file:
        try:
            design = Design.from_toml(design_file)
        except DesignValidationError as e:
            log.critical("%s", e)
            sys.exit(1)
    else:
        if not xeda_project:
            sys.exit(
                "No design file or project files were specified and no `xedaproject.toml` was found in the working directory."
            )
        designs = xeda_project.designs
        if not designs:
            log.critical(
                "There are no designs in the xedaproject file. You can specify a single design description using `--design-file` argument."
            )
            sys.exit(1)
        assert isinstance(xeda_project.design_names, list)  # type checker
        log.info(
            "Available designs in xedaproject: %s",
            ", ".join(xeda_project.design_names),
        )
        design = xeda_project.get_design(design_name)
        if not design:
            if design_name:
                log.critical(
                    'Design "%s" not found in %s. Available designs are: %s',
                    design_name,
                    xedaproject,
                    ", ".join(xeda_project.design_names),
                )
                sys.exit(1)
            else:
                if len(designs) == 1:
                    design = designs[1]
                else:
                    if console.is_interactive:
                        terminal_menu = TerminalMenu(
                            xeda_project.design_names, title="Please select a design: "
                        )
                        idx = terminal_menu.show()
                        if idx is None or not isinstance(idx, int) or idx < 0:
                            sys.exit("Invalid design choice!")
                        design = designs[idx]
                    else:
                        design_name = click.prompt(
                            "Please enter design name: ",
                            type=click.Choice(xeda_project.design_names),
                        )
                        if (
                            not design_name
                            or design_name not in xeda_project.design_names
                        ):
                            sys.exit("Invalid design name!")
                        design = xeda_project.get_design(design_name)
            if not design:
                sys.exit(
                    "[ERROR] no design was specified and none were automatically discovered."
                )

    flow_overrides = settings_to_dict(flow_settings)
    log.debug("flow_overrides: %s", flow_overrides)
    flow_overrides = {**flows_config.get(flow, {}), **flow_overrides}
    flow_class = discover_flow_class(flow)
    runner = DefaultRunner(
        xeda_run_dir,
        debug=options.debug,
        cached_dependencies=cached_dependencies,
        run_in_existing_dir=run_in_existing_dir,
    )
    try:
        runner.run_flow(flow_class, design, flow_overrides)
    except FlowFatalError as e:
        log.critical(
            "Flow %s failed: FlowFatalException %s",
            flow,
            " ".join(str(a) for a in e.args),
        )
        sys.exit(1)
    except NonZeroExitCode as e:
        log.critical(
            "Flow %s failed: NonZeroExitCode %s", flow, " ".join(str(a) for a in e.args)
        )
        sys.exit(1)
    except ExecutableNotFound as e:
        log.critical(
            "Executable '%s' was not found! (tool:%s, flow:%s, PATH:%s)",
            e.exec,
            e.tool,
            flow,
            e.path,
        )
        sys.exit(1)
    except FlowSettingsError as e:
        log.critical("%s", e)
        if options.debug:
            raise e from None
        sys.exit(1)
    except FlowException as e:  # any flow exception
        log.critical("%s", e)
        if options.debug:
            raise e from None
        sys.exit(1)


@cli.command(context_settings=CONTEXT_SETTINGS, short_help="List available flows.")
def list_flows():
    table = Table(
        title="Available flows",
        show_header=True,
        header_style="bold yellow",
        title_style=Style(frame=True, bold=True),
        box=box.HEAVY_HEAD,
        show_lines=True,
    )
    table.add_column("Flow", header_style="bold green", style="bold")
    table.add_column("Description")
    table.add_column("Class", style="dim")
    super_flow_doc = inspect.getdoc(Flow)
    for cls_name, (_mod, cls) in sorted(registered_flows.items()):
        doc = inspect.getdoc(cls)
        if doc == super_flow_doc:
            doc = "<no description>"
        table.add_row(
            cls_name,
            doc,
            str(cls.__module__) + "." + cls.__name__,
        )
    console.print(table)


@cli.command(
    context_settings=CONTEXT_SETTINGS, short_help="List flow settings information"
)
@click.argument(
    "flow",
    metavar="FLOW_NAME",
    type=click.Choice(sorted(list(registered_flows))),
    required=True,
)
@click.pass_context
def list_settings(ctx: click.Context, flow):
    print_flow_settings(flow, options=ctx.obj)


@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Design-space exploration: run several instances of a flow to find optimal parameters and results",
)
@click.argument("flow")
@click.option(
    "--max-cpus",
    default=max(1, multiprocessing.cpu_count()),
    type=int,
    help="Maximum total number of logical CPU cores to use.",
    show_default=True,
    show_envvar=True,
)
@click.pass_context
def dse(ctx, flow, max_cpus):  # pylint: disable=unused-argument
    """Design-space exploration (e.g. fmax)"""


SHELLS: dict[str, dict[str, Any]] = {
    "bash": {
        "eval_file": "~/.bashrc",
        "eval": 'eval "$(_XEDA_COMPLETE=bash_source xeda)"',
        "completion_file": None,
    },
    "zsh": {
        "eval_file": "~/.zshrc",
        "eval": 'eval "$(_XEDA_COMPLETE=zsh_source xeda)"',
        "completion_file": None,
    },
    "fish": {
        "eval_file": "~/.config/fish/completions/xeda.fish",
        "eval": "eval (env _XEDA_COMPLETE=fish_source xeda)",
        "completion_file": None,
    },
}


@cli.command(
    context_settings=CONTEXT_SETTINGS,
    short_help="Shell completion",
    help=f"Name of the shell. Supported shells are: {', '.join(SHELLS.keys())}",
)
@click.argument(
    "shell",
    required=False,
    type=click.Choice(list(SHELLS.keys()), case_sensitive=False),
)
@click.option(
    "--stdout",
    is_flag=True,
    cls=ClickMutex,
    help="""Produce the shell completion script and output it to the standard output.\n
    Example usage:\n
         $ xeda completion zsh --stdout  > $HOME/.zsh/completion/_xeda\n
         $ xeda completion bash --stdout > $HOME/.local/share/bash-completion/xeda
    """,
)
@click.pass_context
def completion(_ctx: click.Context, stdout, shell=None):
    """Xeda shell auto-completion"""
    os_default_shell_name = None
    os_default_shell = os.environ.get("SHELL")
    if os_default_shell:
        os_default_shell_name = os_default_shell.split(os.sep)[-1]
    if os_default_shell_name:
        if not shell:
            shell = os_default_shell_name
        elif os_default_shell_name != shell and not stdout:
            console.print(
                f"[yellow]WARNING:[/] Current default shell ([bold]{os_default_shell}[/]) is different from the specified shell [bold]{shell}[/b]"
            )
    assert shell is not None
    if stdout:
        completion_class = get_completion_class(shell)
        if completion_class:
            complete = completion_class(
                cli=cli, ctx_args={}, prog_name=__package__, complete_var="source_xeda"
            )
            print(complete.source())
    else:
        shell_desc = SHELLS.get(shell, {})
        console.print(
            f"""
    Make sure 'xeda' executable is properly installed and is accessible through the shell's PATH.
        [yellow]$ xeda --version[/]
    Add the following line to [magenta underline]{shell_desc.get('eval_file')}[/]:
        {shell_desc.get('eval')}
            """,
            highlight=False,
        )


@cli.command()
@click.argument("subcommand", required=False)
@click.pass_context
def help(ctx: click.Context, subcommand=None):  # pylint: disable=redefined-builtin
    if subcommand:
        subcommand_obj = cli.get_command(ctx, subcommand)
        if subcommand_obj is None:
            click.echo("I don't know that command.")
        else:
            click.echo(subcommand_obj.get_help(ctx))
    else:
        click.echo(cli.get_usage(ctx))



import logging
from typing import Any, Dict

from ..flow import FpgaSynthFlow
from ...dataclass import validator
from . import Vivado
from .vivado_synth import RunOptions, VivadoSynth

log = logging.getLogger(__name__)

# curated options based on experiments and:
# see https://www.xilinx.com/support/documentation/sw_manuals/xilinx2020_1/ug904-vivado-implementation.pdf
# and https://www.xilinx.com/support/documentation/sw_manuals/xilinx2020_1/ug901-vivado-synthesis.pdf
xeda_strategies: Dict[str, Dict[str, Any]] = {
    "Debug": {
        "synth": {
            "-assert": None,
            "-debug_log": None,
            "-flatten_hierarchy": "none",
            "-keep_equivalent_registers": None,
            "-no_lc": None,
            "-fsm_extraction": "off",
            "-directive": "RuntimeOptimized",
        },
        "opt": {"-directive": "RuntimeOptimized"},
        "place": {"-directive": "RuntimeOptimized"},
        "place_opt": {},
        "route": {"-directive": "RuntimeOptimized"},
        "phys_opt": {"-directive": "RuntimeOptimized"},
    },
    "Runtime": {
        "synth": {"-directive": "RuntimeOptimized"},
        "opt": {"-directive": "RuntimeOptimized"},
        "place": {"-directive": "RuntimeOptimized"},
        "place_opt": {},
        # with -ultrathreads results are not reproducible!
        # OR "-no_timing_driven -ultrathreads",
        "route": {"-directive": "RuntimeOptimized"},
        "phys_opt": {"-directive": "RuntimeOptimized"},
    },
    "Default": {
        "synth": {"-flatten_hierarchy": "rebuilt", "-directive": "Default"},
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "Default"},
        "place_opt": {},
        "route": {"-directive": "Default"},
        "phys_opt": {"-directive": "Default"},
    },
    "Timing": {
        # -mode: default, out_of_context
        # -flatten_hierarchy: rebuilt, full; equivalent in terms of QoR?
        # -no_lc: When checked, this option turns off LUT combining
        # -keep_equivalent_registers -no_lc
        "synth": {
            "-flatten_hierarchy": "rebuilt",
            "-retiming": None,
            "-directive": "PerformanceOptimized",
            "-fsm_extraction": "one_hot",
            #   "-resource_sharing off",
            #   "-no_lc",
            "-shreg_min_size": "5",
            #   "-keep_equivalent_registers "
        },
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "ExtraPostPlacementOpt"},
        "place_opt": {
            "-retarget": None,
            "-propconst": None,
            "-sweep": None,
            "-aggressive_remap": None,
            "-shift_register_opt": None,
        },
        "phys_opt": {"-directive": "AggressiveExplore"},
        "place_opt2": {"-directive": "Explore"},
        # "route": "-directive NoTimingRelaxation",
        "route": {"-directive": "AggressiveExplore"},
    },
    "ExtraTimingCongestion": {
        "synth": {
            "-flatten_hierarchy": "full",
            "-retiming": None,
            "-directive": "PerformanceOptimized",
            "-fsm_extraction": "one_hot",
            "-resource_sharing": "off",
            "-shreg_min_size": "10",
            "-keep_equivalent_registers": None,
        },
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "AltSpreadLogic_high"},
        "place_opt": {
            "-retarget": None,
            "-propconst": None,
            "-sweep": None,
            "-remap": None,
            "-muxf_remap": None,
            "-aggressive_remap": None,
            "-shift_register_opt": None,
        },
        "place_opt2": {"-directive": "Explore"},
        "phys_opt": {"-directive": "AggressiveExplore"},
        "route": {"-directive": "AlternateCLBRouting"},
    },
    "ExtraTiming": {
        "synth": {
            "-flatten_hierarchy": "full",
            "-retiming": None,
            "-directive": "PerformanceOptimized",
            "-fsm_extraction": "one_hot",
            "-resource_sharing": "off",
            #   "-no_lc",
            "-shreg_min_size": "10",
            "-keep_equivalent_registers": None,
        },
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "ExtraTimingOpt"},
        "place_opt": {
            "-retarget": None,
            "-propconst": None,
            "-sweep": None,
            "-muxf_remap": None,
            "-aggressive_remap": None,
            "-shift_register_opt": None,
        },
        "place_opt2": {"-directive": "Explore"},
        "phys_opt": {"-directive": "AggressiveExplore"},
        "route": {"-directive": "NoTimingRelaxation"},
    },
    "ExtraTimingAltRouting": {
        # -mode: default, out_of_context
        # -flatten_hierarchy: rebuilt, full; equivalent in terms of QoR?
        # -no_lc: When checked, this option turns off LUT combining
        # -keep_equivalent_registers -no_lc
        "synth": {
            "-flatten_hierarchy": "full",
            "-retiming": None,
            "-directive": "PerformanceOptimized",
            "-fsm_extraction": "one_hot",
            #   "-resource_sharing off",
            #   "-no_lc",
            "-shreg_min_size": "5",
            "-keep_equivalent_registers": None,
        },
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "ExtraTimingOpt"},
        "place_opt": {
            "-retarget": None,
            "-propconst": None,
            "-sweep": None,
            "-aggressive_remap": None,
            "-shift_register_opt": None,
        },
        "phys_opt": {"-directive": "AggressiveExplore"},
        # "route": "-directive NoTimingRelaxation",
        "route": {"-directive": "AlternateCLBRouting"},
    },
    "Area": {
        # AreaOptimized_medium or _high prints error messages in Vivado 2020.1: "unexpected non-zero reference counts", but succeeeds and post-impl sim is OK too
        "synth": {
            "-flatten_hierarchy": "full",
            "-control_set_opt_threshold": "1",
            "-shreg_min_size": "3",
            "-resource_sharing": "auto",
            "-directive": "AreaOptimized_medium",
        },
        # if no directive: -resynth_seq_area
        "opt": {"-directive": "ExploreArea"},
        "place": {"-directive": "Default"},
        "place_opt": {"-directive": "ExploreArea"},
        # "place_opt": {'-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
        #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'},
        # if no directive: -placement_opt
        "phys_opt": {"-directive": "Explore"},
        "route": {"-directive": "Explore"},
    },
    "AreaHigh": {
        # AreaOptimized_medium or _high prints error messages in Vivado 2020.1: "unexpected non-zero reference counts", but succeeeds and post-impl sim is OK too
        "synth": {
            "-flatten_hierarchy": "full",
            "-control_set_opt_threshold": "1",
            "-shreg_min_size": "3",
            "-resource_sharvng": "on",
            "-directive": "AreaOptimized_high",
        },
        # if no directive: -resynth_seq_area
        "opt": {"-directive": "ExploreArea"},
        "place": {"-directive": "Default"},
        "place_opt": {"-directive": "ExploreArea"},
        # "place_opt": {'-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
        #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'},
        # if no directive: -placement_opt
        "phys_opt": {"-directive": "Explore"},
        "route": {"-directive": "Explore"},
    },
    "AreaPower": {
        # AreaOptimized_medium or _high prints error messages in Vivado 2020.1: "unexpected non-zero reference counts", but succeeeds and post-impl sim is OK too
        "synth": {
            "-flatten_hierarchy": "full",
            "-control_set_opt_threshold": "1",
            "-shreg_min_size": "3",
            "-resource_sharing": "auto",
            "-gated_clock_conversion": "auto",
            "-directive": "AreaOptimized_medium",
        },
        "opt": {"-directive": "ExploreArea"},
        "place": {"-directive": "Default"},
        "place_opt": {
            "-retarget": None,
            "-propconst": None,
            "-sweep": None,
            "-aggressive_remap": None,
            "-shift_register_opt": None,
            "-dsp_register_opt": None,
            "-resynth_seq_area": None,
            "-merge_equivalent_drivers": None,
        },
        "place_opt2": {"-directive": "ExploreArea"},
        # FIXME!!! This is the only option that results in correct post-impl timing sim! Why??!
        "phys_opt": {"-directive": "AggressiveExplore"},
        "route": {"-directive": "Explore"},
    },
    "AreaTiming": {
        "synth": {"-flatten_hierarchy": "rebuilt", "-retiming": None},
        # if no directive: -resynth_seq_area
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "ExtraPostPlacementOpt"},
        # "place_opt": {"-directive ExploreArea"},
        "place_opt": {
            "-retarget": None,
            "-propconst": None,
            "-sweep": None,
            "-aggressive_remap": None,
            "-shift_register_opt": None,
            "-dsp_register_opt": None,
            "-resynth_seq_area": None,
            "-merge_equivalent_drivers": None,
        },
        "place_opt2": {"-directive": "ExploreArea"},
        # if no directive: -placement_opt
        "phys_opt": {"-directive": "AggressiveExplore"},
        "route": {"-directive": "Explore"},
    },
    "AreaExploreWithRemap": {
        "synth": {"-flatten_hierarchy": "full", "-retiming": None},
        # if no directive: -resynth_seq_area
        "opt": {"-directive": "ExploreWithRemap"},
        "place": {"-directive": "Default"},
        "place_opt": {"-directive": "ExploreWithRemap"},
        # "place_opt": {'-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
        #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'},
        # if no directive: -placement_opt
        "phys_opt": {"-directive": "Explore"},
        "route": {"-directive": "Explore"},
    },
    "AreaExploreWithRemap2": {
        "synth": {"-flatten_hierarchy": "full", "-retiming": None},
        # if no directive: -resynth_seq_area
        "opt": {"-directive": "ExploreArea"},
        "place": {"-directive": "Default"},
        "place_opt": {"-directive": "ExploreWithRemap"},
        # "place_opt": {'-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
        #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'},
        # if no directive: -placement_opt
        "phys_opt": {"-directive": "Explore"},
        "route": {"-directive": "Explore"},
    },
    "AreaExplore": {
        "synth": {"-flatten_hierarchy": "full"},
        # if no directive: -resynth_seq_area
        "opt": {"-directive": "ExploreArea"},
        "place": {"-directive": "Default"},
        "place_opt": {"-directive": "ExploreArea"},
        # "place_opt": {'-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
        #               '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'},
        # if no directive: -placement_opt
        "phys_opt": {"-directive": "Explore"},
        "route": {"-directive": "Explore"},
    },
    "Power": {
        "synth": {
            "-flatten_hierarchy": "full",
            "-gated_clock_conversion": "auto",
            "-control_set_opt_threshold": "1",
            "-shreg_min_size": "3",
            "-resource_sharing": "auto",
        },
        # if no directive: -resynth_seq_area
        "opt": {"-directive": "ExploreSequentialArea"},
        "place": {"-directive": "Default"},
        "place_opt": {"-directive": "ExploreSequentialArea"},
        # {'-retarget', '-propconst', '-sweep', '-aggressive_remap', '-shift_register_opt',
        #   '-dsp_register_opt', '-bram_power_opt', '-resynth_seq_area', '-merge_equivalent_drivers'},
        # if no directive: -placement_opt
        "phys_opt": {"-directive": "Explore"},
        "route": {"-directive": "Explore"},
    },
}


def _vivado_steps(strategy: str, s: str) -> Dict[str, Any]:
    if s == "synth":
        steps = ["synth", "opt", "power_opt"]
    else:
        steps = [
            "place",
            "power_opt",
            "place_opt",
            "place_opt2",
            "phys_opt",
            "phys_opt2",
            "route",
        ]
    return {step: xeda_strategies[strategy].get(step) for step in steps}


class VivadoAltSynth(Vivado, FpgaSynthFlow):
    """Synthesize with Xilinx Vivado using an alternative TCL-based flow"""

    class Settings(VivadoSynth.Settings):
        synth: RunOptions = RunOptions(
            strategy="Default", steps=_vivado_steps("Default", "synth")
        )
        impl: RunOptions = RunOptions(
            strategy="Default", steps=_vivado_steps("Default", "impl")
        )

        @validator("synth")
        def validate_synth(cls, v: RunOptions):
            if v.strategy and not v.steps:
                v.steps = _vivado_steps(v.strategy, "synth")
            return v

        @validator("impl", always=True)
        def validate_impl(cls, v: RunOptions, values):
            if not v.strategy:
                synth = values.get("synth")
                if synth:
                    v.strategy = synth.strategy
            if v.strategy:
                v.steps = {**_vivado_steps(v.strategy, "impl"), **v.steps}
            return v

    def run(self):
        clock_xdc_path = self.copy_from_template("clock.xdc")
        ss = self.settings
        assert isinstance(ss, self.Settings)

        if ss.out_of_context:
            if "synth" in ss.synth.steps and ss.synth.steps["synth"] is not None:
                ss.synth.steps["synth"]["-mode"] = "out_of_context"

        blacklisted_resources = ss.blacklisted_resources
        if "bram" in blacklisted_resources and "bram_tile" not in blacklisted_resources:
            blacklisted_resources.append("bram_tile")

        log.info("blacklisted_resources: %s", blacklisted_resources)

        # if 'bram_tile' in blacklisted_resources:
        #     # FIXME also add -max_uram 0 for ultrascale+
        #     settings.synth.steps['synth'].append('-max_bram 0')
        # if 'dsp' in blacklisted_resources:
        #     settings.synth.steps['synth'].append('-max_dsp 0')

        self.add_template_filter(
            "flatten_dict",
            lambda d: " ".join(
                [f"{k} {v}" if v is not None else k for k, v in d.items()]
            ),
        )

        script_path = self.copy_from_template(
            "vivado_alt_synth.tcl",
            xdc_files=[clock_xdc_path],
        )
        self.vivado.run("-source", script_path)



"""Utilities for command line interface"""
import logging
import re
import sys
from dataclasses import dataclass
from functools import reduce
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import click
from click_help_colors import HelpColorsGroup
from overrides import overrides
from rich import box
from rich.style import Style
from rich.table import Table
from rich.text import Text

from .console import console
from .dataclass import asdict
from .flow_runner import FlowNotFoundError, get_flow_class
from .flows.flow import Flow
from .utils import set_hierarchy, try_convert

__all__ = [
    "ClickMutex",
    "OptionEatAll",
    "ConsoleLogo",
    "XedaHelpGroup",
    "discover_flow_class",
]


log = logging.getLogger(__name__)


@dataclass
class XedaOptions:
    verbose: bool = False
    quiet: bool = False
    debug: bool = False


class ClickMutex(click.Option):
    """Mutual exclusion of options"""

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive_with: list = kwargs.pop("mutually_exclusive_with", [])
        self.required_if: list = kwargs.pop("required_if", [])
        if self.mutually_exclusive_with:
            kwargs["help"] = (
                kwargs.get("help", "")
                + "Option is mutually exclusive with "
                + ", ".join(self.mutually_exclusive_with)
                + "."
            ).strip()
        if self.required_if:
            kwargs["help"] = (
                kwargs.get("help", "")
                + "Option is required only if "
                + " or ".join(self.mutually_exclusive_with)
                + " is specified "
            ).strip()
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.name in opts:
            for mutex_opt in self.mutually_exclusive_with:
                if mutex_opt in opts:
                    raise click.UsageError(
                        f"Option {self.name} is mutually exclusive with {mutex_opt}."
                    )
        else:
            for dependent_opt in self.required_if:
                if dependent_opt in opts:
                    raise click.UsageError(
                        f"Option {self.name} is required when {dependent_opt} is specified."
                    )

        return super().handle_parse_result(ctx, opts, args)


class OptionEatAll(click.Option):
    """
    Taken from https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click#answer-48394004.
    """  # noqa: ignore=E501 pylint: disable=C0301

    @overrides
    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop("save_other_options", True)
        super().__init__(*args, **kwargs)
        self._previous_parser_process: Optional[
            Callable[[Union[str, Tuple[str, ...]], click.parser.ParsingState], None]
        ] = None
        self._eat_all_parser = None

    @overrides
    def add_to_parser(self, parser, ctx):
        def parser_process(value: str, state: click.parser.ParsingState):
            """method to hook to the parser.process"""
            value_list = [value]
            if self._eat_all_parser is not None and self.save_other_options:
                # grab everything up to the next option
                done = False
                while state.rargs:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                            break
                    if done:
                        break
                    value_list.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value_list += state.rargs
                state.rargs[:] = []

            # call the actual process
            if self._previous_parser_process is not None:
                self._previous_parser_process(tuple(value_list), state)

        super().add_to_parser(parser, ctx)
        for name in self.opts:
            # pylint: disable=protected-access
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process  # type: ignore
                our_parser.process = parser_process
                break


class ConsoleLogo:
    """Xeda ASCII logo on console"""

    logo: str = """
:==-+-:        :=-.-='.---==:::::::==-. .--=:::::::::=.          ,+==::::::+=-.
 .= {X}O{O} -=     .=--.==. *  ,...........:: *:  {X}........{O}  `=-      ,='            *
   :=:--=-  =-..-=.   +  .-:::::::::::' +: {X}=+++++++++-{O} `:+   ,=' .*########:  *
     -=.-:=-.-:=:     +  +:             +: {X}=+++++++++*.{O} `=- .=: :*########%:  *
       -==-.-=-       +  +: ,:::::::-,  +: {X}=++++++++++*:{O} := =: :*%########%:  *
      .==-.-==-       +  +: `:::::::='  +: {X}=++++++++++*:{O} := =: `*%#######%*'  *
     -= -:==. ==-     +  +:             +: {X}=++++++++++*:{O} .= =:                *
   := -==-  ==. ==-   *  `+..........-. +. {X}=+++++++++*:{O} .+  =: .==::::::::=.  *
 .:-.==:     .=- {X}O{O} =. :=+-............: *, {X}:++++++++'{O} .+-   =: -=          +  *
:=--==.        :=--.=: `=+:::::::::::-' `*-::::::::::+-'    `=-='          `-='
    """

    @classmethod
    def print(cls):
        """print the logo if console is wide enough"""
        if console.width >= 80:
            console.print(
                cls.logo.format(X="[dark_orange]", O="[/]"),
                highlight=False,
                emoji=False,
            )


class XedaHelpGroup(HelpColorsGroup):
    """How to display CLI help"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.help_headers_color = "yellow"
        self.help_options_color = "green"

    def format_usage(self, ctx: click.Context, formatter: click.HelpFormatter):
        ConsoleLogo.print()
        super().format_usage(ctx, formatter)


# DictStrHier = Dict[str, "StrOrDictStrHier"]
DictStrHier = Dict[str, Any]
StrOrDictStrHier = Union[str, DictStrHier]


def settings_to_dict(
    settings: Union[
        None, List[str], Tuple[str, ...], Dict[str, StrOrDictStrHier], Flow.Settings
    ],
) -> Dict[str, Any]:
    if not settings:
        return {}
    if isinstance(settings, (tuple, list)):
        res: DictStrHier = {}
        for override in settings:
            sp = override.split("=")
            if len(sp) != 2:
                raise ValueError("Settings should be in KEY=VALUE format!")
            key, val = sp
            set_hierarchy(res, key, try_convert(val, convert_lists=True))
        return res
    if isinstance(settings, Flow.Settings):
        return asdict(settings)
    if isinstance(settings, dict):
        return settings
    raise TypeError(f"overrides is of unsupported type: {type(settings)}")


def print_flow_settings(flow, options: XedaOptions):
    flow_class = discover_flow_class(flow)
    schema = flow_class.Settings.schema(by_alias=True)
    type_defs = {}

    def get_type(field):
        if isinstance(field, (list, tuple)):
            return Text("|".join(str(get_type(f)) for f in field))
        typ = field.get("type")
        ref = field.get("$ref")
        if typ is None and ref:
            typ = ref.split("/")[-1]
            typ_def = schema.get("definitions", {}).get(typ)
            if typ not in type_defs:
                type_defs[typ] = typ_def
            elif type_defs[typ] != typ_def:
                log.critical(
                    "type definition for %s changed!\nPrevious def:\n %s new def:\n %s",
                    typ,
                    type_defs[typ],
                    typ_def,
                )
            return Text(typ, style="blue")
        additional = field.get("additionalProperties")
        if typ == "object" and additional:
            return Text(f"Dict[string -> {get_type(additional)}]")

        if typ == "array":
            items = field.get("items")
            if items:
                return Text(f"array[{get_type(items)}]")

        def join_types(lst, joiner, style="red"):
            return reduce(
                lambda x, y: x + Text(joiner, style) + y, (get_type(t) for t in lst)
            )

        allof = field.get("allOf")
        if allof:
            return join_types(allof, " & ")  # intersection
        anyOf = field.get("anyOf")
        if anyOf:
            return join_types(anyOf, " | ")  # union
        return Text(typ)

    table = Table(
        title=f"{flow} settings",
        show_header=True,
        header_style="bold yellow",
        title_style=Style(frame=True, bold=True),
        box=box.HEAVY_HEAD,
        show_lines=True,
    )
    table.add_column("Property", header_style="bold green", style="bold")
    table.add_column("Type", max_width=32)
    table.add_column("Default", max_width=42)
    table.add_column("Description")

    def fmt_default(v: Any) -> str:
        if isinstance(v, str):
            return f'"{v}"'
        if isinstance(v, bool):
            return str(v).lower()
        return str(v)

    for name, field in schema.get("properties", {}).items():
        if name in ["results"] or name in Flow.Settings.__fields__.keys():
            continue
        required = name in schema.get("required", [])
        desc: str = field.get("description", "")
        typ = get_type(field)
        req_or_def = (
            "[red]<required>[/red]" if required else fmt_default(field.get("default"))
        )
        table.add_row(name, typ, req_or_def, desc)
    console.print(table)

    for typ_name, typ_def in list(type_defs.items()):
        for prop, prop_def in typ_def.get("properties").items():
            get_type(prop_def)
    for typ_name, typ_def in list(type_defs.items()):
        c = "blue"
        table = Table(
            title=f"Type [{c}]{typ_name}[/{c}]",
            show_header=True,
            header_style="bold green",
            title_style=Style(frame=True, bold=True),
            box=box.SQUARE_DOUBLE_HEAD,
            show_lines=True,
        )
        if options.debug:
            console.print(f"Type: {typ_name}")
            console.print_json(data=typ_def)
        table.add_column("property")
        table.add_column("type")
        table.add_column("description")
        for prop, prop_def in typ_def.get("properties").items():
            if prop_def.get("hidden_from_schema"):
                continue
            if (
                typ_name.endswith("__Settings")
                and prop in Flow.Settings.__fields__.keys()
            ):
                continue
            desc = prop_def.get("description", prop_def.get("title", "-"))
            desc = re.sub(r"\s*\.*\s*$", "", desc)
            table.add_row(prop, get_type(prop_def), desc)
        console.print(table)


def discover_flow_class(flow: str) -> Type[Flow]:
    try:
        return get_flow_class(flow, "xeda.flows", __package__)
    except FlowNotFoundError:
        log.critical(
            "Flow %s is not known to Xeda. Please make sure the name is correctly specified.",
            flow,
        )
        sys.exit(1)



import html
import logging
from typing import Any, Dict
from xml.etree import ElementTree

from .vivado_sim import VivadoSim
from .vivado_postsynthsim import VivadoPostsynthSim
from .vivado_synth import VivadoSynth

logger = logging.getLogger(__name__)


class VivadoPower(VivadoSim):
    class Settings(VivadoSim.Settings):
        timing_sim: bool = True
        elab_debug: str = "typical"
        saif: str = "activity.saif"
        postsynthsim: VivadoPostsynthSim.Settings
        power_report_xml: str = "power_impl_timing.xml"

    def init(self) -> None:
        assert self.design.tb, "A testbench is required for power estimation"
        ss = self.settings
        assert isinstance(ss, self.Settings)
        ss.postsynthsim.timing_sim = ss.timing_sim
        ss.postsynthsim.elab_debug = ss.elab_debug
        ss.postsynthsim.saif = ss.saif
        self.add_dependency(VivadoPostsynthSim, ss.postsynthsim)

    def run(self) -> None:
        assert isinstance(self.settings, self.Settings)

        postsynth_sim_flow = self.pop_dependency(VivadoPostsynthSim)
        synth_flow = postsynth_sim_flow.pop_dependency(VivadoSynth)

        checkpoint = str(
            synth_flow.run_path
            / f"{self.design.name}.runs"
            / "impl_1"
            / f"{self.design.rtl.top}_routed.dcp"
        )
        saif_file = str(postsynth_sim_flow.run_path / self.settings.saif)

        # assert isinstance(dep_synth_flow.settings, VivadoSynth.Settings)
        script_path = self.copy_from_template(
            "vivado_power.tcl",
            checkpoint=checkpoint,
            saif_file=saif_file,
        )

        self.vivado.run("-source", script_path)

    def parse_power_report(self, report_xml) -> Dict[str, Any]:
        tree = ElementTree.parse(report_xml)
        results = {}
        for tablerow in tree.findall("./section[@title='Summary']/table/tablerow"):
            tablecells = tablerow.findall("tablecell")
            key, value = (
                html.unescape(x.attrib["contents"]).strip() for x in tablecells
            )
            results[key] = value

        for tablerow in tree.findall(
            "./section[@title='Summary']/section[@title='On-Chip Components']/table/tablerow"
        ):
            tablecells = tablerow.findall("tablecell")
            if len(tablecells) >= 2:
                contents = [
                    html.unescape(x.attrib["contents"]).strip() for x in tablecells
                ]
                key = contents[0]
                value = contents[1]
                results[f"Component Power: {key}"] = value

        return results

    def parse_reports(self) -> bool:
        assert isinstance(self.settings, self.Settings)
        report_xml = self.run_path / self.settings.power_report_xml
        results = self.parse_power_report(report_xml)
        self.results.update(**results)
        return True



from importlib.metadata import version, PackageNotFoundError
from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = version("xeda")
except PackageNotFoundError:
    # package is not installed
    try:
        __version__ = get_distribution("xeda").version
    except DistributionNotFound:
        # package is not installed
        pass
